from importlib.metadata import version

VERSION_PROG = version("pybirdsreynolds")
SHORT_DESCRIPTION = (
    "pybirdsreynolds is an interactive simulation of the Reynolds Boids model."
)
MARGIN = 1
WIDTH_CONTROLS_DEFAULT = 70
WIDTH_PARAMS_DEFAULT = 350
HEIGHT_PARAMS_CONTROLS_DEFAULT = 600
FONT_TYPE_LIST = [
    "Nota Sans Mono",
    "Consolas",
    "Courier",
    "Ubuntu Sans Mono",
    "DejaVu Sans Mono",
    "Courier New",
    "Liberation Mono",
    "Monaco", 
    "Menlo",
    "PT Mono",
    "Fira Mono", 
    "JetBrains Mono",
    "Source Code Pro", 
    "Inconsolata",
]
