PAUSED = True
BLINK_STATE = True
FPS = False
HIDDEN = False
TRANS_HIDDEN = False
SHIFT_PRESSED = False
REPEATING = {"active": False, "job": None}
CANVAS_BG = None
FILL_COLOR = None
OUTLINE_COLOR = None
INFO_COLOR = None
WIDTH_CONTROLS = 70
WIDTH_PARAMS = 350
HEIGHT_PARAMS_CONTROLS = 600
FRAME_COUNT = 0
FPS_VALUE = 0
LAST_TIME = 0
COUNT = 0
POINTS_ID = []
WIDTH_BEFORE_MAXIMIZED = 0
HEIGHT_BEFORE_MAXIMIZED = 0
TIP_WINDOW = None
AVG_X = 0
AVG_Y = 0
AVG_SPEED_ABS = 0
AVG_VX = 0
AVG_VY = 0
AVERAGE = False
HUNTER = False
